/* Simple button daemon
 *
 * Petr Malat (c) 2008-2009 
 * http://malat.biz/
 *
 * Distributed under the terms of GNU GPL version 3 
 * */

#define _GNU_SOURCE
#include <stdio.h>
#include <string.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/time.h>
#include <fcntl.h>
#include <linux/input.h>
#include <stdlib.h>
#include <unistd.h>
#include <stdarg.h>
#include <errno.h>
#include <getopt.h>
#include <signal.h>

/** Maximum number of actions read from configuration file */
#define  MAX_ACTIONS 16
/** Array with intervals and associated commands */
struct   action_s { float min; float max; char *cmd; } actions[MAX_ACTIONS];
/** Number of actions */
unsigned actions_nmemb;
/** Log is written to this file. If it's set to NULL, loging is disabled */
FILE*    logger_file;
/** Filedescriptor for open event device */
int      event_fd;
/** Pid is writen to file with this name */
char*    pid_fn;


/** Structure with default values - this value is used if it's not overiden on command line */
struct {
	char * const conf;
	char * const dev;
	char * const logger;
	char * const pid;
} const  defaults = {"/etc/btnd.conf", "/dev/input/event0", NULL, "/var/run/btnd.pid"};

/** Levels of importance for logger() */
enum logger_e { 
	/** Fatal error, causes daemon to exit */
	FATAL, 
	/** Noncritical error, the daemon can run */
	ERROR, 
	/** Information */
	INFO 
};

/** Writes message to log. 
 * @param level - Importance level of message
 * @param cmd   - format string as in printf()
 * @param ...   - arguments to format
 */
void logger(enum logger_e level, char *cmd, ...){
	if( logger_file != NULL ){
		va_list list;
		va_start(list, cmd);
		vfprintf(logger_file, cmd, list);
		va_end(list);
	}
	if(level == FATAL)
		exit(EXIT_FAILURE);
}

/** Executes command */
void system_exec(const char* cmd){
	pid_t pid;

	pid = fork();
	if( pid == 0 ){ /* child */
		logger(INFO, "Executing '%s'.\n", cmd);
		if( system(cmd) == -1){
			logger(ERROR, "Couldn't execute '%s'. %s.\n", cmd, strerror(errno));
			exit(EXIT_FAILURE);
		} else {
			logger(INFO, "Completed '%s'.\n", cmd);
			exit(EXIT_SUCCESS);
		}
	} else if (pid == -1){ 
		logger(FATAL, "Couldn't fork new process. %s.\n", strerror(errno));
	}
}

/** Proccess events */
int process_events(int fd){
	struct input_event event;
	time_t KEY_POWER_sec = 0;
	suseconds_t KEY_POWER_usec = 0;
	float sec;
	int i;

	while(1){
		read(fd, &event, sizeof(event));
		if( event.type == EV_KEY ){
			switch(event.code){
				case KEY_POWER:
					if(event.value == 1){
						KEY_POWER_sec = event.time.tv_sec;
						KEY_POWER_usec = event.time.tv_usec;
					} else {
						sec = ((event.time.tv_sec - KEY_POWER_sec)*1000000 + event.time.tv_usec - KEY_POWER_usec)/1000000.;
						for(i = 0; i < sizeof(actions)/sizeof(struct action_s); i++){
							if( sec > actions[i].min && sec < actions[i].max )
								system_exec(actions[i].cmd);
						}
					}
			}
		}
	}
	return 0;
}

/** Compare two intervals, used when sorting actions array */
int actions_cmp(const void *pa, const void *pb)
{
	const struct action_s *a = (struct action_s *)pa, *b = (struct action_s *)pb;
	if(a->min < b->min) return -1;
	if(a->min > b->min) return 1;
	return 0;
}

/** Reads configuration file and sets up actions array */
void parse_config(FILE *conf)
{
	float min, max;
	char *cmd, c;
	actions_nmemb = 0;

	while(!feof(conf)){
		if(fscanf(conf, " %c", &c) == 1){
			if(c == '#'){
				fscanf(conf, "%*[^\n]\n");
				continue;
			} else {
				ungetc(c, conf);
				if(fscanf(conf, "%f%*[ \t]%f%*[ \t]%a[^\n]\n", &min, &max, &cmd) == 3){
					if(! (min < max)){
						logger(FATAL, "Semantic error in configuration file: "
							"Interval start is not less than interval end (%f, %f).\n", min, max);
					}
					if(actions_nmemb < MAX_ACTIONS){
						actions[actions_nmemb].min = min;
						actions[actions_nmemb].max = max;
						actions[actions_nmemb].cmd = cmd;
						actions_nmemb++;
					} else {
						logger(FATAL, "Maximum number of actions (%d) reached. Ignoring following actions.\n", 
								MAX_ACTIONS);
						break;
					}
					continue;
				}
			}	
		}
		logger(FATAL, "Syntax error in configuration file.\n");
	}

	qsort(actions, actions_nmemb, sizeof(*actions), actions_cmp);
	int i;
	for(i = 1; i < actions_nmemb; i++)
		if(actions[i-1].max > actions[i].min){
			logger(FATAL, "Semantic error in configuration file: Intervals (%f %f) and (%f %f) intersect.\n", 
				actions[i-1].min, actions[i-1].max, actions[i].min, actions[i].max);
		}
}

/** Prints help */
void print_help(void){
	fprintf(stderr, "btnd 0.2   Build: " __DATE__ " " __TIME__ "\n"
		"Copyright 2008-2009 Petr Malat (http://malat.biz/).\n"
		"This is free software with ABSOLUTELY NO WARRANTY.\n\n");

	fprintf(stderr, "Usage: btnd [-c config_file] [-d device_file] [-l log_file]\n"
			"            [-p pid_file ] [-h]\n"
			"  -c config_file - configuration file, default is %s\n"
			"  -d device_file - event device, default is %s\n"
			"  -l log_file    - log filename, default is %s (no logging)\n"
			"  -p pid_file    - write pid into this file %s\n"
			"  -h             - print this message\n", defaults.conf, defaults.dev, defaults.logger, defaults.pid);
}

/** Handles command line arguments */
void parse_cmd (int argc, char *argv[])
{
	const char *dev_fn = defaults.dev;
	const char *logger_fn = defaults.logger;
	const char *conf_fn = defaults.conf;
	int opt;
	FILE *conf;

	while( (opt = getopt(argc, argv, "c:p:l:d:h")) != -1 ){
		switch(opt){
			case 'c':
				conf_fn = optarg;
				break;
			case 'd':
				dev_fn = optarg;
				break;
			case 'l':
				logger_fn = optarg;
				break;
			case 'p':
				pid_fn = optarg;
				break;
			case ':': 
			case '?': 
				fprintf(stderr, "Bad commandline syntax\n");
			case 'h':
			default:
				print_help();
				exit(0);
		}
	}

	/* Open logger */
	if( logger_fn != NULL ){
		if( (logger_file = fopen(logger_fn, "a")) == NULL){
			fprintf(stderr, "Couldn't open logger '%s'. %s\n", logger_fn, strerror(errno));
			exit(EXIT_FAILURE);
		}
	}

	/* Open configuration file */
	if((conf = fopen(conf_fn, "r")) == NULL)
		logger(FATAL, "Couldn't open configuration file '%s'. %s\n", conf_fn, strerror(errno));
	parse_config(conf);
	fclose(conf);

	/* Open device */
	if( (event_fd = open(dev_fn, O_RDONLY)) == -1)
		logger(FATAL, "Couldn't open device '%s'. %s.\nDefault device can be created with mknod %s c 13 64\n",
				dev_fn, strerror(errno), defaults.dev);
}

int main(int argc, char *argv[]){
	struct sigaction sa;
	FILE *pid_file;
	pid_t pid;

	parse_cmd(argc, argv);

	pid = fork();
	if( pid == 0 ){
		close(0); close(1); close(2);
		sa.sa_handler = SIG_IGN;
		sigaction(SIGCHLD, &sa, NULL);
		sigaction(SIGTSTP, &sa, NULL);
		sigaction(SIGTTOU, &sa, NULL);
		sigaction(SIGTTIN, &sa, NULL);
		setsid();
		logger(INFO, "Daemon succesfully started.\n"); 
		return process_events(event_fd);
	} else if (pid == -1){
		logger(FATAL, "Couldn't fork daemon process. %s.\n", strerror(errno));
	}
	if((pid_file = fopen(pid_fn, "w")) == NULL){
		logger(ERROR, "Couldn't open pid file '%s'. %s.\n", pid_fn, strerror(errno)); 
	} else {
		fprintf(pid_file, "%d\n", pid);
		fclose(pid_file);
	}
	return 0;
}

